#include <stdio.h>
#ifdef _MSC_VER
#include <windows.h>
#else
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#endif
#include "i_net.h"

#define MAXSERVERS 512

#define mst_server 100
#define mst_end    101

int     nowtime;
buf_t   launcher_message;
byte    launcher_message_buffer[MAX_UDP_PACKET];

typedef struct
{
    netadr_t   address;
    int        last_received;   // for timeouts
    bool       active;

    char       map[32];	
    char       wadname[128];
    byte       wad;
    byte       maxclients;
    byte       clients;
} server_t;

server_t *servers;
int       numservers;

#ifdef _MSC_VER
int I_GetTime (void)
{
	static DWORD  basetime;
	DWORD         tm;

	tm = timeGetTime();
	if (!basetime)
		basetime = tm;

	return (tm-basetime)/1000;
}

#else

// Returns time in seconds
int I_GetTime(void)
{
        struct timeval tv;
        struct timezone tz;
        long long int thistimereply;
	static long long int basetime;	

        gettimeofday(&tv, &tz);

        thistimereply = tv.tv_sec * 1000 + tv.tv_usec / 1000;

	if (!basetime)
            basetime = thistimereply;
        
	return (thistimereply - basetime)/1000;
}

void daemon_init(void)
{
    int   pid;
    FILE *fpid;
    
    if ((pid = fork()) != 0)
	exit(0);
	
    pid = getpid();
    fpid = fopen("MasterPID", "w");
    fprintf(fpid, "PID %d\n", pid);	
    fclose(fpid);
}
#endif

int FindServerByAddress(void)
{
	int i;

	for (i=0; i<MAXSERVERS; i++)
	{
		if (!servers[i].active)
			continue;

		if (NET_CompareAdr(servers[i].address, net_from) )
			return i;
	}

	for (i=0; i<MAXSERVERS; i++)
		if (!servers[i].active)
		{
			servers[i].active = true;
			servers[i].address = net_from;
			return i;
		}
	
	return -1;
}


void ParseCommands(void)
{
	int        sv;
	int        challenge;
	server_t  *server;

	challenge = MSG_ReadLong();
		
	if (challenge == SERVER_CHALLENGE)
	{
		sv = FindServerByAddress();
		if (sv == -1)
		{
			printf ("Error: too many servers\n");
			return;
		}

		server = &servers[sv];
		server->last_received = nowtime;		
		printf ("%s\n", NET_AdrToString(server->address));
	}
	
	if (challenge == LAUNCHER_CHALLENGE)
	{
		SZ_Clear(&launcher_message);
		MSG_WriteLong(&launcher_message, LAUNCHER_CHALLENGE);

		for (int i=0; i<MAXSERVERS; i++)
		{
			server_t *usv = &servers[i];
		
			if (!usv->active)
				continue;

			MSG_WriteByte(&launcher_message, mst_server);			
			MSG_WriteString(&launcher_message, NET_AdrToString (usv->address) );
	    }

	    MSG_WriteByte(&launcher_message, mst_end);

	    NET_SendPacket(launcher_message.cursize, launcher_message.data, net_from);
	}
}


void CheckTimeouts(void)
{
	for (int i=0; i<MAXSERVERS; i++)
	{
		if (!servers[i].active)
			continue;

		if ( (nowtime - servers[i].last_received) >= 60)
			servers[i].active = false;
	}

}


//
// csDoom Master Server MAIN() function
//
int main()
{
	localport = MASTERPORT;
	InitNetCommon();

	servers = new server_t[MAXSERVERS];
	memset (servers, 0, sizeof(server_t)*MAXSERVERS);
	numservers = 0;

        launcher_message.data = launcher_message_buffer;
	launcher_message.maxsize = sizeof(launcher_message_buffer);

	printf ("csDoom master server initialized.\n");
	
#ifndef _MSC_VER
	daemon_init();
#endif
	
	while (1)
	{
		nowtime = I_GetTime();
		I_DoSelect();
		
		while (NET_GetPacket() )
			ParseCommands();

		CheckTimeouts();
	}	
	
//	delete[] servers;
//	CloseNetwork();

	return 0;
}